package parse;

import org.xml.sax.Attributes;

import xcore.List;
import xcore.Str;
import xcore.Value;
import env.Env;

public class Element extends Parser {

  // An XML Element. For this to succeed the current XML input element
  // must have the required tag, the attribute values are bound to the
  // variable names and all the children must match the child parser.

  private String tag;
  private Parser child;
  private List<Str> names;

  public String getTag() {
	return tag;
  }

  public void setTag(String tag) {
	this.tag = tag;
  }

  public Parser getChild() {
	return child;
  }

  public void setChild(Parser child) {
	this.child = child;
  }

  public List<Str> getNames() {
	return names;
  }

  public void setNames(List<Str> names) {
	this.names = names;
  }

  public Element(Str tag, List<Str> names) {
	this(tag, new OK(), names);
  }

  public Element(Str tag, Parser child, List<Str> names) {
	super();
	this.tag = tag.getValue();
	this.child = child;
	this.names = names;
  }

  public TagProcessor parse(final Grammar grammar, final Success s, final Env<String, Value> env, final Fail c, final Fail f) {

	// Return a TagProcessor that expects a SAX event for the
	// expected tag. If that tag occurs then return a TagProcessor
	// for the children; when the child parser succeeds, expect an
	// end tag SAX event for the parent. If any of this fails then
	// just bail out using the current fail continuation...

	return new TagProcessor() {
	  public TagProcessor processTag(String tag, final Attributes attributes, Boolean isStart) {
		grammar.tab(System.out);
		grammar.debug("Element " + Element.this.tag);
		grammar.indent();

		// Expect the start tag...
		if (tag.equals(Element.this.tag) && isStart) {
		  Success newSucc = new Success() {
			public TagProcessor success(final Object value, final Env<String, Value> env, Fail fail) {
			  return new TagProcessor() {
				// Expect the end tag...
				public TagProcessor processTag(String tag, Attributes ignore, Boolean isStart) {
				  if (tag.equals(Element.this.tag) && !isStart) {
					grammar.undent();
					grammar.tab(System.out);
					grammar.debug("Element /" + Element.this.tag + " found.");
					return s.success(value, env, c);
				  } else
					throw new Error("Expecting /" + tag);
				}
			  };
			}
		  };
		  Env<String, Value> newEnv = env;
		  for (Str name : names) {
			Str value = new Str(attributes.getValue(name.getValue()));
			grammar.tab(System.out);
			grammar.debug("BIND " + name.getValue() + " = " + value);
			newEnv = newEnv.bind(name.getValue(), value);
		  }
		  // In the mean-time the child parser must be satisfied...
		  return child.parse(grammar, newSucc, newEnv, c, c);
		} else {
		  // Otherwise we did not get what we expected so just
		  // opt out using the current fail...
		  grammar.undent();
		  grammar.tab(System.out);
		  grammar.debug("FAIL Element " + Element.this.tag + " not found(" + tag + ")");
		  return f.fail(tag, attributes, isStart);
		}
	  }
	};
  }

  public String toString() {
	return "Element(" + tag + "," + child + "," + names + ")";
  }

}
