package parse;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

import xcore.List;
import xcore.Str;
import xcore.Value;
import env.Env;
import env.NullEnv;

public class XMLReader extends DefaultHandler {

  // An XMLReader is used to process an XML document with respect
  // to a grammar. The reader handles the SAX events with respect
  // to the state of a parse using TagProcessors...

  private InputStream in; // Source of XML.
  private Value value; // The synthesized value.
  private TagProcessor processor; // Will handle the next event.
  private Parser language; // The starting parser.
  private Grammar grammar; // The grammar.

  private Env<String, Value> env = new NullEnv<String, Value>();

  public XMLReader(Grammar grammar, String startingNT, List<Value> values, String filename) {
	this.grammar = grammar;
	Rule rule = grammar.findRule(startingNT, values.size());
	if (rule == null)
	  throw new Error("No rule " + startingNT + "/" + values.size());
	List<Str> args = rule.getArgs();
	for (int i = 0; i < Math.min(args.size(), values.size()); i++)
	  env = env.bind(args.get(i).getValue(), values.get(i));
	this.language = rule.getBody();
	File file = new File(filename);
	if (file.exists())
	  try {
		this.in = new FileInputStream(file);
	  } catch (FileNotFoundException e) {
		throw new Error("File " + filename + " does not exist.");
	  }
	else
	  throw new Error("File " + filename + " does not exist.");
  }

  public void characters(char[] chars, int start, int length) {
  }

  public void endElement(String uri, String localName, String qname) {
	
	// At this point, the child-part of a grammar must be satisfied and
	// we are expecting an end tag that matches the original start tag...
	
	processor = processor.processTag(qname, null, false);
  }

  public void error(SAXParseException e) {
	System.out.println("SAX Parse Error: " + e.toString());
  }

  public Value parse() {
	try {
	  SAXParserFactory factory = SAXParserFactory.newInstance();
	  SAXParser parser = factory.newSAXParser();
	  Success success = new Success() {
		public TagProcessor success(Object value, Env<String, Value> env, Fail fail) {
		  XMLReader.this.setValue((Value) value);
		  return null;
		}
	  };
	  Fail fail = new Fail() {
		public TagProcessor fail(String tag, Attributes attributes, Boolean isStart) {
		  throw new Error("FAIL");
		}
	  };
	  Fail cut = fail;
	  processor = language.parse(grammar, success, env, cut, fail);
	  parser.parse(in, this);
	  in.close();
	  return value;
	} catch (org.xml.sax.SAXParseException x) {
	  x.printStackTrace();
	  try {
		in.close();
	  } catch (IOException ioe) {
		ioe.printStackTrace();
	  }
	} catch (Throwable e) {
	  e.printStackTrace();
	  try {
		in.close();
	  } catch (IOException e1) {
		e1.printStackTrace();
	  }
	}
	return null;
  }

  protected void setValue(Value value) {
	this.value = value;
  }

  public void startElement(String uri, String localName, String qname, Attributes attributes) {
	
	// Each time a SAX element event occurs, use the current processor
	// to process the tag and attributes. This will return a new 
	// TagProcessor that should be used for the children...
	
	processor = processor.processTag(qname, attributes, true);
  }

  public void warning(SAXParseException e) {
	System.out.println(e.toString());
  }
}